__doc__ = """
REHO - A Comprehensive Decision Support Tool for Sustainable Energy System Planning.
====================================================================================
Renewable Energy Hub Optimizer (REHO) is a decision support tool for sustainable urban energy system planning.
It is developed by EPFL (Switzerland), within the Industrial Process and Energy Systems Engineering (IPESE) group.
"""