__doc__ = """
Contains plotting functions and code relative to the visualization of REHO results.

- ``layout.csv``: Contains colors and labels to characterize the units and layers of an energy system configuration.
- ``sia380_1.csv``: Contains the translation of building's affectation in roman numbering to labels in the SIA 380/1 norm.
"""
